@@ -20,10 +20,12 @@
 import os
 import re
 import sys
+import tempfile
 import unittest
+import warnings
 from StringIO import StringIO
 
-sys.path.insert(0, '../debian_bundle/')
+sys.path.insert(0, '../lib/debian/')
 
 import deb822
 
@@ -146,7 +148,7 @@ CcYAoOLYDF5G1h3oR1iDNyeCI6hRW03S
     ]
 
 
-CHANGES_FILE = '''\
+CHANGES_FILE = u'''\
 Format: 1.7
 Date: Fri, 28 Dec 2007 17:08:48 +0100
 Source: bzr-gtk
@@ -235,6 +237,47 @@ iD8DBQFIGWQO0UIZh3p4ZWERAug/AJ93DWD9o+1VMgPDjWn/dsmPSgTWGQCeOfZi
 -----END PGP SIGNATURE-----
 '''
 
+UNPARSED_PARAGRAPHS_WITH_COMMENTS = '''\
+# Leading comments should be ignored.
+
+Source: foo
+Section: bar
+# An inline comment in the middle of a paragraph should be ignored.
+Priority: optional
+Homepage: http://www.debian.org/
+
+# Comments in the middle shouldn't result in extra blank paragraphs either.
+
+# Ditto.
+
+# A comment at the top of a paragraph should be ignored.
+Package: foo
+Architecture: any
+Description: An awesome package
+  # This should still appear in the result.
+  Blah, blah, blah. # So should this.
+# A comment at the end of a paragraph should be ignored.
+
+# Trailing comments shouldn't cause extra blank paragraphs.
+'''
+
+PARSED_PARAGRAPHS_WITH_COMMENTS = [
+    deb822.Deb822Dict([
+        ('Source', 'foo'),
+        ('Section', 'bar'),
+        ('Priority', 'optional'),
+        ('Homepage', 'http://www.debian.org/'),
+    ]),
+    deb822.Deb822Dict([
+        ('Package', 'foo'),
+        ('Architecture', 'any'),
+        ('Description', 'An awesome package\n'
+            '  # This should still appear in the result.\n'
+            '  Blah, blah, blah. # So should this.'),
+    ]),
+]
+
+
 class TestDeb822Dict(unittest.TestCase):
     def make_dict(self):
         d = deb822.Deb822Dict()
@@ -341,7 +384,8 @@ class TestDeb822(unittest.TestCase):
             self.assertWellParsed(deb822_, PARSED_PACKAGE)
 
     def test_gpg_info(self):
-        if not os.path.exists('/usr/bin/gpgv'):
+        if not (os.path.exists('/usr/bin/gpgv') and
+                os.path.exists('/usr/share/keyrings/debian-keyring.gpg')):
             return
 
         unparsed_with_gpg = SIGNED_CHECKSUM_CHANGES_FILE % CHECKSUM_CHANGES_FILE
@@ -351,16 +395,25 @@ class TestDeb822(unittest.TestCase):
         result_from_file = deb822_from_file.get_gpg_info()
         deb822_from_lines = deb822.Dsc(unparsed_with_gpg.splitlines())
         result_from_lines = deb822_from_lines.get_gpg_info()
-        valid = {'GOODSIG':  ['D14219877A786561', 'John Wright <john.wright@hp.com>'],
-                 'VALIDSIG': ['8FEFE900783CF175827C2F65D14219877A786561', '2008-05-01',
-                              '1209623566', '0', '3', '0', '17', '2', '01',
-                              '8FEFE900783CF175827C2F65D14219877A786561'],
-                 'SIG_ID':   ['mQFnUWWR1Gr6itMV7Bx5L4N60Wo', '2008-05-01', '1209623566']}
+
+        valid = {
+         'GOODSIG':  ['D14219877A786561', 'John Wright <john.wright@hp.com>'],
+         'VALIDSIG': ['8FEFE900783CF175827C2F65D14219877A786561', '2008-05-01',
+                      '1209623566', '0', '3', '0', '17', '2', '01',
+                      '8FEFE900783CF175827C2F65D14219877A786561'],
+         'SIG_ID':   ['j3UjSpdky92fcQISbm8W5PlwC/g', '2008-05-01',
+                      '1209623566'],
+        }
 
         for result in result_from_str, result_from_file, result_from_lines:
-            self.assertEqual(len(result.keys()), len(valid.keys()))
-            for k,v in valid.items():
-                self.assertEqual(''.join(v), ''.join(result[k]))
+            # The second part of the GOODSIG field could change if the primary
+            # uid changes, so avoid checking that.  Also, the first part of the
+            # SIG_ID field has undergone at least one algorithm changein gpg,
+            # so don't bother testing that either.
+            self.assertEqual(set(result.keys()), set(valid.keys()))
+            self.assertEqual(result['GOODSIG'][0], valid['GOODSIG'][0])
+            self.assertEqual(result['VALIDSIG'], valid['VALIDSIG'])
+            self.assertEqual(result['SIG_ID'][1:], valid['SIG_ID'][1:])
 
     def test_iter_paragraphs_array(self):
         text = (UNPARSED_PACKAGE + '\n\n\n' + UNPARSED_PACKAGE).splitlines()
@@ -385,10 +438,10 @@ class TestDeb822(unittest.TestCase):
                 self.assertWellParsed(d, PARSED_PACKAGE)
             self.assertEqual(count, 2)
 
-    def _test_iter_paragraphs(self, file, cls, **kwargs):
+    def _test_iter_paragraphs(self, filename, cls, **kwargs):
         """Ensure iter_paragraphs consistency"""
         
-        f = open(file)
+        f = open(filename)
         packages_content = f.read()
         f.close()
         # XXX: The way multivalued fields parsing works, we can't guarantee
@@ -398,10 +451,12 @@ class TestDeb822(unittest.TestCase):
 
         s = StringIO()
         l = []
-        for p in cls.iter_paragraphs(open(file), **kwargs):
+        f = open(filename)
+        for p in cls.iter_paragraphs(f, **kwargs):
             p.dump(s)
             s.write("\n")
             l.append(p)
+        f.close()
         self.assertEqual(s.getvalue(), packages_content)
         if kwargs["shared_storage"] is False:
             # If shared_storage is False, data should be consistent across
@@ -629,6 +684,143 @@ Description: python modules to work with Debian-related data formats
             d3['Some-Test-Key'] = 'some value'
         self.assertEqual(d3.dump(), "Some-Test-Key: some value\n")
 
+    def test_unicode_values(self):
+        """Deb822 objects should contain only unicode values
+
+        (Technically, they are allowed to contain any type of object, but when
+        parsed from files, and when only string-type objects are added, the
+        resulting object should have only unicode values.)
+        """
+
+        objects = []
+        objects.append(deb822.Deb822(UNPARSED_PACKAGE))
+        objects.append(deb822.Deb822(CHANGES_FILE))
+        objects.extend(deb822.Deb822.iter_paragraphs(file('test_Packages')))
+        objects.extend(deb822.Packages.iter_paragraphs(file('test_Packages')))
+        objects.extend(deb822.Deb822.iter_paragraphs(file('test_Sources')))
+        objects.extend(deb822.Deb822.iter_paragraphs(
+                         file('test_Sources.iso8859-1'), encoding="iso8859-1"))
+        for d in objects:
+            for value in d.values():
+                self.assert_(isinstance(value, unicode))
+
+        # The same should be true for Sources and Changes except for their
+        # _multivalued fields
+        multi = []
+        multi.append(deb822.Changes(CHANGES_FILE))
+        multi.append(deb822.Changes(SIGNED_CHECKSUM_CHANGES_FILE
+                                    % CHECKSUM_CHANGES_FILE))
+        multi.extend(deb822.Sources.iter_paragraphs(file('test_Sources')))
+        for d in multi:
+            for key, value in d.items():
+                if key.lower() not in d.__class__._multivalued_fields:
+                    self.assert_(isinstance(value, unicode))
+
+    def test_encoding_integrity(self):
+        utf8 = list(deb822.Deb822.iter_paragraphs(file('test_Sources')))
+        latin1 = list(deb822.Deb822.iter_paragraphs(
+                                                file('test_Sources.iso8859-1'),
+                                                encoding='iso8859-1'))
+
+        # dump() with no fd returns a unicode object - both should be identical
+        self.assertEqual(len(utf8), len(latin1))
+        for i in range(len(utf8)):
+            self.assertEqual(utf8[i].dump(), latin1[i].dump())
+
+        # XXX: The way multiline fields parsing works, we can't guarantee
+        # that trailing whitespace is reproduced.
+        utf8_contents = "\n".join([line.rstrip() for line in
+                                   file('test_Sources')] + [''])
+        latin1_contents = "\n".join([line.rstrip() for line in
+                                     file('test_Sources.iso8859-1')] + [''])
+
+        utf8_to_latin1 = StringIO()
+        for d in utf8:
+            d.dump(fd=utf8_to_latin1, encoding='iso8859-1')
+            utf8_to_latin1.write("\n")
+
+        latin1_to_utf8 = StringIO()
+        for d in latin1:
+            d.dump(fd=latin1_to_utf8, encoding='utf-8')
+            latin1_to_utf8.write("\n")
+
+        self.assertEqual(utf8_contents, latin1_to_utf8.getvalue())
+        self.assertEqual(latin1_contents, utf8_to_latin1.getvalue())
+
+    def test_mixed_encodings(self):
+        """Test that we can handle a simple case of mixed encodings
+
+        In general, this isn't guaranteed to work.  It uses the chardet
+        package, which tries to determine heuristically the encoding of the
+        text given to it.  But as far as I've seen, it's reliable for mixed
+        latin1 and utf-8 in maintainer names in old Sources files...
+        """
+
+        # Avoid spitting out the encoding warning during testing.
+        warnings.filterwarnings(action='ignore', category=UnicodeWarning)
+
+        filename = 'test_Sources.mixed_encoding'
+        for paragraphs in [deb822.Sources.iter_paragraphs(file(filename)),
+                           deb822.Sources.iter_paragraphs(file(filename),
+                                                          use_apt_pkg=False)]:
+            p1 = paragraphs.next()
+            self.assertEqual(p1['maintainer'],
+                             u'Adeodato Simó <dato@net.com.org.es>')
+            p2 = paragraphs.next()
+            self.assertEqual(p2['uploaders'],
+                             u'Frank Küster <frank@debian.org>')
+
+    def test_bug597249_colon_as_first_value_character(self):
+        """Colon should be allowed as the first value character. See #597249.
+        """
+
+        data = 'Foo: : bar'
+        parsed = {'Foo': ': bar'}
+        self.assertWellParsed(deb822.Deb822(data), parsed)
+
+    @staticmethod
+    def _dictset(d, key, value):
+        d[key] = value
+
+    def test_field_value_ends_in_newline(self):
+        """Field values are not allowed to end with newlines"""
+
+        d = deb822.Deb822()
+        self.assertRaises(ValueError, self._dictset, d, 'foo', 'bar\n')
+        self.assertRaises(ValueError, self._dictset, d, 'foo', 'bar\nbaz\n')
+
+    def test_field_value_contains_blank_line(self):
+        """Field values are not allowed to contain blank lines"""
+
+        d = deb822.Deb822()
+        self.assertRaises(ValueError, self._dictset, d, 'foo', 'bar\n\nbaz')
+        self.assertRaises(ValueError, self._dictset, d, 'foo', '\n\nbaz')
+
+    def test_multivalued_field_contains_newline(self):
+        """Multivalued field components are not allowed to contain newlines"""
+
+        d = deb822.Dsc()
+        # We don't check at set time, since one could easily modify the list
+        # without deb822 knowing.  We instead check at get time.
+        d['Files'] = [{'md5sum': 'deadbeef', 'size': '9605', 'name': 'bad\n'}]
+        self.assertRaises(ValueError, d.get_as_string, 'files')
+
+    def _test_iter_paragraphs_comments(self, paragraphs):
+        self.assertEqual(len(paragraphs), len(PARSED_PARAGRAPHS_WITH_COMMENTS))
+        for i in range(len(paragraphs)):
+            self.assertWellParsed(paragraphs[i],
+                                  PARSED_PARAGRAPHS_WITH_COMMENTS[i])
+
+    def test_iter_paragraphs_comments_use_apt_pkg(self):
+        paragraphs = list(deb822.Deb822.iter_paragraphs(
+            UNPARSED_PARAGRAPHS_WITH_COMMENTS.splitlines(), use_apt_pkg=True))
+        self._test_iter_paragraphs_comments(paragraphs)
+
+    def test_iter_paragraphs_comments_native(self):
+        paragraphs = list(deb822.Deb822.iter_paragraphs(
+            UNPARSED_PARAGRAPHS_WITH_COMMENTS.splitlines(), use_apt_pkg=False))
+        self._test_iter_paragraphs_comments(paragraphs)
+
 
 class TestPkgRelations(unittest.TestCase):
 
@@ -719,7 +911,8 @@ class TestPkgRelations(unittest.TestCase):
                         'version': None}],
                     [{'name': 'autoconf', 'version': None, 'arch': None}],
                     [{'name': 'debhelper', 'version': ('>>', '5.0.0'), 'arch': None}]],
-                'build-depends-indep': []}
+                'build-depends-indep': [],
+                'binary': [[{'name': 'apache2-mpm-itk', 'version': None, 'arch': None}]]}
         self.assertEqual(rel1, pkg1.relations)
         pkg2 = pkgs.next()
         rel2 = {'build-conflicts': [],
@@ -737,8 +930,88 @@ class TestPkgRelations(unittest.TestCase):
                     [{'name': 'file', 'version': None, 'arch': None}],
                     [{'name': 'bzip2', 'version': None, 'arch': None}],
                     [{'name': 'lsb-release', 'version': None, 'arch': None}]],
-                'build-depends-indep': []}
+                'build-depends-indep': [],
+                'binary': [[{'name': 'binutils', 'version': None, 'arch': None}],
+                    [{'name': 'binutils-dev', 'version': None, 'arch': None}],
+                    [{'name': 'binutils-multiarch', 'version': None, 'arch': None}],
+                    [{'name': 'binutils-hppa64', 'version': None, 'arch': None}],
+                    [{'name': 'binutils-spu', 'version': None, 'arch': None}],
+                    [{'name': 'binutils-doc', 'version': None, 'arch': None}],
+                    [{'name': 'binutils-source', 'version': None, 'arch': None}]]}
         self.assertEqual(rel2, pkg2.relations)
 
+
+class TestGpgInfo(unittest.TestCase):
+
+    def setUp(self):
+        # These tests can only run with gpgv and a keyring available.  When we
+        # can use Python >= 2.7, we can use the skip decorator; for now just
+        # check in each test method whether we should actually run.
+        self.should_run = (
+            os.path.exists('/usr/bin/gpgv') and
+            os.path.exists('/usr/share/keyrings/debian-keyring.gpg'))
+
+        self.data = SIGNED_CHECKSUM_CHANGES_FILE % CHECKSUM_CHANGES_FILE
+        self.valid = {
+            'GOODSIG':
+                ['D14219877A786561', 'John Wright <john.wright@hp.com>'],
+            'VALIDSIG':
+                ['8FEFE900783CF175827C2F65D14219877A786561', '2008-05-01',
+                 '1209623566', '0', '3', '0', '17', '2', '01',
+                 '8FEFE900783CF175827C2F65D14219877A786561'],
+            'SIG_ID':
+                ['j3UjSpdky92fcQISbm8W5PlwC/g', '2008-05-01', '1209623566'],
+        }
+
+    def _validate_gpg_info(self, gpg_info):
+        # The second part of the GOODSIG field could change if the primary
+        # uid changes, so avoid checking that.  Also, the first part of the
+        # SIG_ID field has undergone at least one algorithm changein gpg,
+        # so don't bother testing that either.
+        self.assertEqual(set(gpg_info.keys()), set(self.valid.keys()))
+        self.assertEqual(gpg_info['GOODSIG'][0], self.valid['GOODSIG'][0])
+        self.assertEqual(gpg_info['VALIDSIG'], self.valid['VALIDSIG'])
+        self.assertEqual(gpg_info['SIG_ID'][1:], self.valid['SIG_ID'][1:])
+
+    def test_from_sequence_string(self):
+        if not self.should_run:
+            return
+
+        gpg_info = deb822.GpgInfo.from_sequence(self.data)
+        self._validate_gpg_info(gpg_info)
+
+    def test_from_sequence_newline_terminated(self):
+        if not self.should_run:
+            return
+
+        sequence = StringIO(self.data)
+        gpg_info = deb822.GpgInfo.from_sequence(sequence)
+        self._validate_gpg_info(gpg_info)
+
+    def test_from_sequence_no_newlines(self):
+        if not self.should_run:
+            return
+
+        sequence = self.data.splitlines()
+        gpg_info = deb822.GpgInfo.from_sequence(sequence)
+        self._validate_gpg_info(gpg_info)
+
+    def test_from_file(self):
+        if not self.should_run:
+            return
+
+        fd, filename = tempfile.mkstemp()
+        fp = os.fdopen(fd, 'w')
+        fp.write(self.data)
+        fp.close()
+
+        try:
+            gpg_info = deb822.GpgInfo.from_file(filename)
+        finally:
+            os.remove(filename)
+
+        self._validate_gpg_info(gpg_info)
+
+
 if __name__ == '__main__':
     unittest.main()
